module Main exposing (main)

import Browser
import Browser.Dom as Dom
import Css
import Css.Global as Global
import Css.Reset as Reset
import Dict exposing (Dict)
import Html.Styled as Html exposing (Html)
import Html.Styled.Attributes as Attributes exposing (css)
import Html.Styled.Events as Events
import Json.Decode as Decode
import Markdown
import Task
import Texts


type alias Model =
    { scrollOffsets : Dict String Float
    , expandoStates : Dict String ExpandoState
    , error : Maybe Dom.Error
    }


type Msg
    = SetExpandoState String ExpandoState
    | SetScroll String Float
    | RestoredScroll (Result Dom.Error ())


init : () -> ( Model, Cmd Msg )
init _ =
    ( { scrollOffsets = Dict.empty
      , expandoStates = Dict.empty
      , error = Nothing
      }
    , Cmd.none
    )


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        SetExpandoState name state ->
            ( { model | expandoStates = Dict.insert name state model.expandoStates }
            , case state of
                Closed ->
                    Cmd.none

                Open ->
                    Dict.get name model.scrollOffsets
                        |> Maybe.withDefault 0
                        |> Dom.setViewportOf (slugify name) 0
                        |> Task.attempt RestoredScroll
            )

        SetScroll name scroll ->
            ( { model | scrollOffsets = Dict.insert name scroll model.scrollOffsets }
            , Cmd.none
            )

        RestoredScroll result ->
            case result of
                Ok _ ->
                    ( model, Cmd.none )

                Err err ->
                    ( { model | error = Just err }, Cmd.none )


view : Model -> Browser.Document Msg
view model =
    { title = "Restoring Scroll Experiment"
    , body =
        List.map Html.toUnstyled
            [ Reset.meyerV2
            , Reset.borderBoxV201408
            , frame
                [ Html.p
                    [ css [ Css.marginBottom (Css.px 10) ] ]
                    [ Html.a
                        [ Attributes.href "https://git.bytes.zone/brian/elm-save-scroll-experiment" ]
                        [ Html.text "Source on git.bytes.zone" ]
                    ]
                , [ ( "Pride and Prejudice", Texts.prideAndPrejudice )
                  , ( "Frankenstein; or, the Modern Prometheus", Texts.frankenstein )
                  ]
                    |> List.map
                        (\( name, text ) ->
                            expando
                                (model.expandoStates |> Dict.get name |> Maybe.withDefault Closed)
                                name
                                (Html.fromUnstyled (Markdown.toHtml [] text))
                        )
                    |> Html.div []
                , Html.p [ css [ Css.marginBottom (Css.px 10) ] ]
                    [ Html.text "Error: "
                    , Html.text (Debug.toString model.error)
                    ]
                , Html.p [ css [ Css.marginBottom (Css.px 10) ] ]
                    [ Html.text "States"
                    , Dict.toList model.expandoStates
                        |> List.map
                            (\( name, state ) ->
                                Html.li [ css [ Css.display Css.block ] ]
                                    [ Html.text name
                                    , Html.text ": "
                                    , case state of
                                        Open ->
                                            Html.text "Open"

                                        Closed ->
                                            Html.text "Closed"
                                    ]
                            )
                        |> Html.ul []
                    ]
                , Html.p [ css [ Css.marginBottom (Css.px 10) ] ]
                    [ Html.text "Positions"
                    , Dict.toList model.scrollOffsets
                        |> List.map
                            (\( name, offset ) ->
                                Html.li [ css [ Css.display Css.block ] ]
                                    [ Html.text name
                                    , Html.text ": "
                                    , Html.text (String.fromFloat offset)
                                    , Html.text "px"
                                    ]
                            )
                        |> Html.ul []
                    ]
                ]
            ]
    }


frame : List (Html msg) -> Html msg
frame =
    Html.main_
        [ css
            [ Css.width (Css.px 800)
            , Css.margin2 (Css.px 20) Css.auto
            , Css.padding (Css.px 20)
            , Css.borderRadius (Css.px 5)
            , Css.backgroundColor (Css.hex "#778899")
            , Css.color (Css.hex "#E4E7EB")
            , Css.fontFamily Css.sansSerif
            , Css.fontSize (Css.px 18)
            ]
        ]


type ExpandoState
    = Open
    | Closed


flipExpandoState : ExpandoState -> ExpandoState
flipExpandoState state =
    case state of
        Open ->
            Closed

        Closed ->
            Open


expando : ExpandoState -> String -> Html Msg -> Html Msg
expando state title contents =
    Html.section
        [ css
            [ Css.margin2 (Css.px 10) Css.zero
            , Css.padding2 (Css.px 10) Css.zero
            , Css.borderRadius (Css.px 5)
            , Css.backgroundColor (Css.hex "#667788")
            ]
        ]
        [ Html.div
            [ css
                [ Css.textAlign Css.center
                , Css.fontSize (Css.em 1.25)
                , Css.paddingBottom (Css.px 5)
                ]
            ]
            [ Html.text title
            , Html.button
                [ css
                    [ Css.float Css.left
                    , Css.marginLeft (Css.px 10)
                    ]
                , Events.onClick (SetExpandoState title (flipExpandoState state))
                ]
                [ Html.text "Toggle" ]
            ]
        , case state of
            Open ->
                Html.div
                    [ Attributes.id (slugify title)
                    , Events.on "scroll"
                        (Decode.float
                            |> Decode.at [ "target", "scrollTop" ]
                            |> Decode.map (SetScroll title)
                        )
                    , css
                        [ Css.maxHeight (Css.px 300)
                        , Css.overflowY Css.scroll
                        , Css.padding (Css.px 10)
                        , Global.descendants
                            [ Global.p
                                [ Css.lineHeight (Css.em 1.25)
                                , Css.marginBottom (Css.px 10)
                                ]
                            ]
                        ]
                    ]
                    [ contents ]

            Closed ->
                Html.text ""
        ]


slugify : String -> String
slugify =
    String.replace " " "-"
        >> String.replace "." ""
        >> String.replace "," ""
        >> String.replace ";" ""


main : Program () Model Msg
main =
    Browser.document
        { init = init
        , update = update
        , view = view
        , subscriptions = \_ -> Sub.none
        }
